/* hexe - ui.c
 Copyright (C) 2010 spinout
   rmo@spinout182.com
   spinout182.com
 Feel free to modify and redestribute as you like,
 so long that all that have contributed are given
 due credit for their contributions and the code that
 you got is just as avalible
 
 If you just want a snippet, feel free to take it.
*/

#include "hexe.h"

char * ctrls[NCTRLS * 2] = {
    "^F", "Find",
    "^X", "Exit",
    "^O", "Open",
    "^S", "Save",
    "^T", "Mode",
    "^G", "Goto",
    "^I", "Insert"
};

void save_yesno()
{
    int c;
    blank_line(maxy-1);
    blank_line(maxy);
    SET_COLOR(WHITE, BLACK);
    mvprintw( maxy-1, 0, "Save changes? (y/n)");
    while((c=getch())==-1) usleep(5000);
    if(c == 'y')
        save();
    else if(c == 'n')
        return;

}

void go_to()
{
    char in[9];
    int taddr;
    
    blank_line(maxy-1);
    blank_line(maxy);
    SET_COLOR(WHITE, BLACK);
    mvprintw( maxy-1, 0, "Enter hexadecimal address:");
    mvprintw( maxy, 0, "0x");
    get_hex_str(in, NULL, 8, maxy, 2);
    sscanf(in, "%X", &taddr);
    if(taddr < max_addr && taddr > -1)
        addr = taddr;
    else
        wait_message("Address out of range");
    
}

void insert()
{
    int c, i, times, diff, inslen=0, total_inslen=0;
    char b[2], in[512], *_in = in;
    FILE *insfile;
    char *insdata;
    
    blank_line(maxy-1);
    blank_line(maxy);
    SET_COLOR(WHITE, BLACK);
    mvprintw( maxy-1, 0, "What do you want to insert?");
    SET_COLOR(BLACK, WHITE);
    mvprintw(maxy, 0, "^F");
    SET_COLOR(WHITE, BLACK);
    mvprintw(maxy, 2, " File ");
    SET_COLOR(BLACK, WHITE);
    mvprintw(maxy, 10, "^S");
    SET_COLOR(WHITE, BLACK);
    mvprintw(maxy, 12, " Chars ");
    SET_COLOR(BLACK, WHITE);
    mvprintw(maxy, 20, "^B");
    SET_COLOR(WHITE, BLACK);
    mvprintw(maxy, 22, " Bytes ");
    SET_COLOR(BLACK, WHITE);
    mvprintw(maxy, 30, "^X");
    SET_COLOR(WHITE, BLACK);
    mvprintw(maxy, 32, " Back ");
    refresh();
    
    while((c=getch()) == -1) usleep(5000);
    
    blank_line(maxy-1);
    blank_line(maxy);
    usleep(50);
    
    switch(c)
    {
    case CTRL('F'): /* File */
        SET_COLOR(WHITE, BLACK);
        mvprintw( maxy-1, 0, "File to insert:");
        get_input(in, NULL, 512, ' ', '~', maxy);
        insfile = fopen(in, "r+b");
        if(insfile == NULL)
        {
            wait_message("Error opening file");
            return;
        }
        
        fseek(insfile, 0, SEEK_END);
        inslen = ftell(insfile);
        if(inslen < 1)
        {
            fclose(insfile);
            wait_message("Empty file");
            return;
        }
        insdata = malloc(inslen);
        if(insdata == NULL)
        {
            fclose(insfile);
            wait_message("Error allocating memory for file to insert");
            return;
        }
        rewind(insfile);
        if(fread(insdata, inslen, 1, insfile) != 1)
        {
            fclose(insfile);
            free(insdata);
            wait_message("Error reading data from file");
            return;
        }
        fclose(insfile);
        break;
    case CTRL('S'): /* String of charachters */
        mvprintw( maxy-1, 0, "String to insert:");
        get_input(in, NULL, 512, ' ', '~', maxy);
        if((inslen = strlen(in)) < 1)
        {
            wait_message("Empty string");
            return;
        }
        insdata = malloc(inslen);
        strcpy(insdata, in);
        break;
    case CTRL('B'): /* Byte string */
        mvprintw( maxy-1, 0, "Byte string to insert:");
        blank_line(maxy);
        get_hex_str(in, NULL, 512, maxy, 0);
        if(strlen(in) & 1)
        {
            wait_message("Byte string must be multiple of 2 other than 0 in length");
        }
        insdata = NULL;
        for(inslen=0;*_in; _in+=2)
        {
            b[0] = *_in;
            b[1] = *(_in+1);
            sscanf(b, "%02x", &c);
            insdata = realloc(insdata, inslen+1);
            insdata[inslen++] = c;
        }
        break;
    case CTRL('X'):
        return;
    default:
        wait_message("Unkown option, returning");
        return;
    }
    
    blank_line(maxy-1);
    blank_line(maxy);
    SET_COLOR(WHITE, BLACK);
    mvprintw( maxy-1, 0, "How many times? (prefix with '0x' if hex)");
    
    get_hex_str(in, NULL, 8, maxy, 0);
    if(!strncmp(in, "0x", 2))
        sscanf(in, "0x%x", &times);
    else
        sscanf(in, "%d", &times);
    if(times<1)
        goto end;
    
    total_inslen = times * inslen;
    
    if(mode & INSERT)
    {
        max_addr += (times * inslen);
        data = realloc(data, max_addr);
        rmemcpy(data+addr+total_inslen, data+addr, max_addr - (addr + total_inslen));
    }
    else
    {
        diff = max_addr - (total_inslen + addr);
        if(diff > 0)
        {
            max_addr+= diff;
            data = realloc(data, max_addr);
        }
    }
    for(i=0;i<times;i++)
        memcpy(data+addr+(inslen*i), insdata, inslen);
    unsaved_changes = 1;
  end:
    free(insdata);
    
    return;
}

void key_handler()
{
    int v;
    char d[]={0,0,0};
    switch(key)
    {
    case KEY_UP:
        if(addr - colsize >= 0) addr -= colsize;
        else addr = 0;
        low=0;
        break;
    case KEY_DOWN:
        if(addr + colsize < max_addr) addr += colsize;
        else addr = max_addr-1;
        low=0;
        break;
    case KEY_RIGHT:
        if(addr < max_addr-1) addr++;
        low=0;
        break;
    case KEY_LEFT:
        if(addr-1 >= 0) addr--;
        else addr = 0;
        low=0;
        break;
    case PG_DOWN:
        if(addr - (addr%colsize) + ((maxy+1) * colsize) < max_addr)
        {
            addr = addr - (addr%colsize) + (maxy * colsize);
            low=0;
        }
        break;
    case PG_UP:
        if(addr - (addr%colsize) - ((maxy+1) * colsize) >=0)
        {
            addr = addr - (addr%colsize) - (maxy * colsize);
            low=0;
        }
        break;
    case END:
        addr = max_addr-1;
        low=0;
        break;
    case CTRL('X'):
        if(unsaved_changes)
            save_yesno();
        quit();
        break;
    case CTRLF:
        search();
        break;
    case CTRL('G'):
        go_to();
        break;
    case CTRL('T'):
        mode++;
        mode&=3;
        low=0;
        break;
    case CTRL('S'):
        if(read_only)
            return;
        save();
        break;
    case CTRL('I'):
        if(read_only)
            return;
        insert();
        break;
    case CTRL('O'):
        if(unsaved_changes)
            save_yesno();
        open();
        break;
    default:
        if(read_only)
            return;
        if(mode & INSERT)
        {
            if(mode&TEXT)
            {
                if(key<='~' && key>=' ')
                {
                    unsaved_changes = 1;
                    data = realloc(data, ++max_addr);
                    memcpy(data + addr + 1, data + addr, max_addr - addr - 1);
                    data[addr++] = key;
                }
            }
            else
            {
                key = toupper(key);
                if((key >= 'A' && key <= 'F') || (key >= '0' && key <= '9'))
                {
                    unsaved_changes = 1;
                    d[0] = key;
                    sscanf(d, "%01x", &v);
                    if(low)
                    {
                        data[addr] = (data[addr] & 0xF0) | v;
                        addr ++;
                    }
                    else
                    {
                        data = realloc(data, ++max_addr);
                        memcpy(data + addr + 1, data + addr, max_addr - addr - 1);
                        data[addr] = (v<<4);
                    }
                    low=~low;
                }
            }
        }
        else
        {
            if(mode&TEXT)
            {
                if(key<='~' && key>=' ')
                {
                    data[addr++] = key;
                }
            }
            else
            {
                key = toupper(key);
                if((key >= 'A' && key <= 'F') || (key >= '0' && key <= '9'))
                {
                    unsaved_changes = 1;
                    d[0] = key;
                    sscanf(d, "%01x", &v);
                    if(low)
                    {
                        data[addr] = (data[addr] & 0xF0) | v;
                        addr ++;
                    }
                    else
                        data[addr] = (v<<4) | (data[addr] & 0xF);
                    low=~low;
                }
            }
            if(addr >= max_addr)
                addr--;
        }
        break;
    }
}

void search()
{
    char in[512], *_in = in;
    unsigned char valsearch[sizeof(in)/2];
    unsigned char *needle;
    char b[3] = {0,0,0};
    int size=0, v, r,rn=0, c, find_prev_flag=0;
    
    hlmin=hlmax=-1;
    
    blank_line(maxy-1);
    blank_line(maxy);
    
    SET_COLOR(WHITE, BLACK);
    mvprintw(maxy-1, 0, "Enter 'v' for hex value or 's' for string:", b);
    move(maxy, 0);
    while((c=getch())==-1) usleep(5000);
    
    blank_line(maxy-1);
    blank_line(maxy);
    SET_COLOR(WHITE, BLACK);
    
    c = tolower(c);
    if(c == 'v')
    {
        mvprintw(maxy-1, 0, "Enter a hexadecimal value");
        get_hex_str(in, NULL, 512, maxy, 0);
        
        for(size=0;*_in; _in+=2)
        {
            b[0] = *_in;
            b[1] = *(_in+1);
            sscanf(b, "%02x", &v);
            valsearch[size++] = v;
        }
        needle = valsearch;
      searchloop:
        while(1)
        {
            if(find_prev_flag)
            {
                rn--;
                r = find_previous(data, needle, hlmin-1, size, 1);
            }
            else
            {
                rn++;
                r = find_next(data, needle, hlmax, size, max_addr, 1);
            }
            if(r >= 0)
            {
                hlmin = addr = r;
                hlmax = hlmin + size;

              print_search_result:
                
                curr_min = (addr+size) - ((addr+size) % colsize);
                
                print_rows();
                
                blank_line(maxy-1);
                blank_line(maxy);
                
                SET_COLOR(WHITE, BLACK);
                mvprintw(maxy-1, 0, "Result %i (%08X)", rn, hlmin);
                SET_COLOR(BLACK, WHITE);
                mvprintw(maxy, 0, "^X");
                SET_COLOR(WHITE, BLACK);
                mvprintw(maxy, 2, " End search ");
                SET_COLOR(BLACK, WHITE);
                mvprintw(maxy, 14, "^N");
                SET_COLOR(WHITE, BLACK);
                mvprintw(maxy, 16, " Next ");
                SET_COLOR(BLACK, WHITE);
                mvprintw(maxy, 22, "^P");
                SET_COLOR(WHITE, BLACK);
                mvprintw(maxy, 24, " Previous ");
                refresh();
                
                while((c=getch()) == -1) usleep(5000);
                switch(c)
                {
                case CTRL('X'):
                    return;
                case CTRL('N'):
                    find_prev_flag=0;
                    break;
                case CTRL('P'):
                    find_prev_flag=1;
                    break;
                }
            }
            else
            {
                find_prev_flag ? rn++ : rn --;
                if(rn > 0)
                {
                    wait_message("No more results");
                    goto print_search_result;
                }
                else
                {
                    wait_message("No results found");
                    return;
                }
            }
        }
    }
    else if(c == 's')
    {
        mvprintw(maxy-1, 0, "Enter a string to search:");
        get_input(in, NULL, 512, ' ', '~', maxy);
        if(!*in) return;
        size=strlen(in);
        needle=(unsigned char*)in;
        goto searchloop;
    }
}

void print_rows()
{
    int y, x, p, v, row_addr, highlight;
    
    if(addr < curr_min)
        curr_min = addr - (addr % colsize);
    
  getcurrmax:
    curr_max = curr_min + (maxy-2) * colsize;
    
    if(addr >= curr_max)
    {
        curr_min+=colsize;
        goto getcurrmax;
    }
    
    for(y=0;y<maxy-2;y++)
    {
        row_addr = curr_min + colsize * y;
        if(addr - (addr%colsize) == row_addr)
            SET_COLOR(BLUE, WHITE);
        else
            SET_COLOR(BLACK, WHITE);
        mvprintw(y+1, 0, "%08X", row_addr);
        for(x=p=0;p<colsize;p++)
        {
            highlight = (row_addr+p >= hlmin && row_addr+p < hlmax) ? 1 : 0;
            if(!(p&bytegroup))
            {
                SET_COLOR(BLACK, BLACK);
                mvaddch(y+1, 2 * p + 8 + x, ' ');
                x++;
            }
            if(row_addr+p >= max_addr)
            {
                (mode & TEXT) ? SET_COLOR(WHITE, WHITE) : SET_COLOR(BLUE, BLUE);
                v=0;
            }
            else if(row_addr+p == addr)
            {
                highlight ? SET_COLOR(BLACK, YELLOW) : (mode & TEXT) ? SET_COLOR(WHITE, BLACK) : SET_COLOR(WHITE, BLUE);
                v=data[row_addr+p];
            }
            else
            {
                highlight ? SET_COLOR(YELLOW, BLACK) : (mode & TEXT) ? SET_COLOR(BLACK, WHITE) : SET_COLOR(BLACK, BLUE);
                v=data[row_addr+p];
            }
            mvprintw(y+1, 2 * p + 8 + x, "%02X", v);
        }
        SET_COLOR(BLACK, BLACK);
        addch(' ');
        for(p=0;p<colsize;p++)
        {
            highlight = (row_addr+p >= hlmin && row_addr+p < hlmax) ? 1 : 0;
            if(row_addr+p >= max_addr)
                (mode & TEXT) ? SET_COLOR(WHITE, WHITE) : SET_COLOR(BLUE, BLUE);
            else if(row_addr+p == addr)
                highlight ? SET_COLOR(GREEN, BLACK) : (mode & TEXT) ? SET_COLOR(WHITE, BLUE) : SET_COLOR(WHITE, BLACK);
            else
                highlight ? SET_COLOR(GREEN, WHITE) : (mode & TEXT) ? SET_COLOR(BLACK, BLUE) : SET_COLOR(BLACK, WHITE);
            move(y+1,  (2 * colsize) + p + 9 + x);
            if(row_addr+p < max_addr && data[row_addr+p] >= ' ' && data[row_addr+p] <= '~')
                addch(data[row_addr+p]);
            else
                addch('.');
        }
    }
}

void
print_ui()
{
    int i,p,y=maxy-1;
    char status[40];
    
    blank_line(0);
    SET_COLOR(BLACK, WHITE);
    mvprintw(0, 0, "File: %s", fn, addr);
    SET_COLOR(WHITE, BLACK);
    if(read_only)
        sprintf(status, "read-only ");
    else
        sprintf(status, "%s | %s ", mode & INSERT ? "insert" : "overwrite", mode & TEXT ? "ascii" : "hex");
    mvprintw(0, maxx-strlen(status)+1, status);
    
    blank_line(y);
    blank_line(y+1);
    
    for(i=p=0; i < NCTRLS * 2; i+=2)
    {
        if(p+strlen(ctrls[i])+2+strlen(ctrls[i+1])>maxx)
        {
            y+=1;
            p=0;
        }
        SET_COLOR(BLACK, WHITE);
        mvprintw(y, p, ctrls[i]);
        p+=strlen(ctrls[i]);
        SET_COLOR(WHITE, BLACK);
        mvprintw(y, p, " %s  ", ctrls[i+1]);
        p+=strlen(ctrls[i+1])+2;
    }
    move(maxy,maxx);
}
